// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/api/access_token_interceptor.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/homeserver.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/room/rooms.dart';
import 'package:matrix_sdk/src/util/chopper.dart';
import 'package:meta/meta.dart';

class User implements Identifiable<UserId> {

  @override
  final UserId id;

  User({@required this.id});
}

/// A user which is authenticated and can send messages, join rooms etc.
class LocalUser extends User {
  final Homeserver homeserver;
  final String accessToken;

  String _syncToken;

  final ChopperClient _chopper;
  ClientService get _client => _chopper.getService();

  LocalUser(this.homeserver,
            this.accessToken)
      : _chopper = createChopper(
        uri: homeserver.uri,
        interceptors: [AccessTokenInterceptor(accessToken)]
      ),
      rooms = Rooms(store: homeserver.store);

  final Rooms rooms;

  // Sync //
  /// Syncs data with the user's [homeserver]. Use the returned [Stream] to
  /// get when the sync has been successful (true) or unsuccessful (false).
  Stream<bool> sync() async* {

    // Get sync token from store
    _syncToken = await homeserver.store?.getSyncToken() ?? _syncToken;

    while(true) {

      final Response response = await _client.sync(since: _syncToken ?? '');

      // Store sync token
      _syncToken = response.body['next_batch'] as String;
      await homeserver.store?.setSyncToken(_syncToken);

      await _processRooms(response);

      yield true;
    }
  }

  Future<void> _processRooms(Response response) async {
    await _processJoinedRooms(response.body['rooms']['join']);
  }

  Future<void> _processJoinedRooms(Map<String, dynamic> joinedRooms) async {
    joinedRooms.forEach((String id, dynamic j) async {
      final roomId = RoomId(id);

      Room room = await rooms[roomId];
      if (room == null) {
        room = JoinedRoom.fromJson(roomId, j);
      }

      room.addEvents(j['state']['events']);

      rooms[roomId] = room;
    });
  }

  // End Sync //
}